# Copyright 2021-2023 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from textwrap import dedent

from kubemarine.core.action import Action
from kubemarine.core.patch import RegularPatch
from kubemarine.core.resources import DynamicResources
from kubemarine.packages import disable_unattended_upgrade


class TheAction(Action):
    def __init__(self) -> None:
        super().__init__("Disable unattended upgrade for important packages")

    def run(self, res: DynamicResources) -> None:
        group = res.cluster().make_group_from_roles(['all'])
        group.call(disable_unattended_upgrade)


class DisableUnattendedUpgrades(RegularPatch):
    def __init__(self) -> None:
        super().__init__("disable_unattended_upgrades")

    @property
    def action(self) -> Action:
        return TheAction()

    @property
    def description(self) -> str:
        return dedent(
            f"""\
            Disable unattended upgrades on debian nodes for important packages.
            """.rstrip()
        )


# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import pathlib
from collections.abc import Hashable
from textwrap import dedent
from typing import List, Dict, Callable, Union, cast

import jsonschema
from ordered_set import OrderedSet

from kubemarine.core import utils, log, errors
from kubemarine.core.cluster import KubernetesCluster, EnrichmentStage, enrichment


@enrichment(EnrichmentStage.LIGHT)
def verify_connections(cluster: KubernetesCluster) -> None:
    inventory = utils.convert_native_yaml(cluster.inventory)
    # Run restricted validation of sections that participate in the enrichment of connections.
    # This is primarily needed for `do`, and `migrate_kubemarine` procedures.
    _verify_inventory_by_schema(cluster, inventory, 'internal/connections')
    context = cluster.context
    args: dict = context['execution_arguments']
    procedure = context["initial_procedure"]
    if args.get('procedure_config') and procedure in ('add_node', 'remove_node'):
        # Run full validation for add_node/remove_node
        # It is currently not necessary to restrict validation to connections only as for inventory.
        procedure_inventory = utils.convert_native_yaml(cluster.procedure_inventory)
        _verify_inventory_by_schema(cluster, procedure_inventory, context["initial_procedure"])


@enrichment(EnrichmentStage.FULL)
def verify_inventory(cluster: KubernetesCluster) -> None:
    inventory = utils.convert_native_yaml(cluster.inventory)
    _verify_inventory_by_schema(cluster, inventory, 'cluster')
    context = cluster.context
    args: dict = context['execution_arguments']
    # Skip validation if procedure inventory is optional and not provided
    if args.get('procedure_config'):
        procedure_inventory = utils.convert_native_yaml(cluster.procedure_inventory)
        _verify_inventory_by_schema(cluster, procedure_inventory, context["initial_procedure"])


def _verify_inventory_by_schema(cluster: KubernetesCluster, inventory: dict, schema_name: str) -> None:
    for_procedure = ("" if schema_name == 'cluster'
                     else " for connections" if schema_name == 'internal/connections'
                     else f" for procedure '{schema_name}'")

    root_schema_resource = f'resources/schemas/{schema_name}.json'
    root_schema_path = utils.get_internal_resource_path(root_schema_resource)
    root_schema = pathlib.Path(root_schema_path)
    if not root_schema.exists():
        raise Exception(f"Failed to find schema to validate the inventory file{for_procedure}.")

    with utils.open_internal(root_schema_resource) as f:
        schema = json.load(f)

    validator_cls = jsonschema.validators.validator_for(schema)
    validator_cls.check_schema(schema)

    root_schema_uri = root_schema.as_uri()
    resolver = jsonschema.RefResolver(base_uri=root_schema_uri, referrer=schema)
    validator = validator_cls(schema, resolver=resolver)

    errs = list(validator.iter_errors(inventory))
    if not errs:
        return

    context = cluster.context
    logger = cluster.log
    logger.error(f"Inventory file{for_procedure} is failed to be validated against the schema.")

    errs = _resolve_errors(errs)
    for err in errs:
        logger.verbose("------------------------------------------")
        logger.verbose(_verbose_msg(validator, err))

    debug_filepath = log.get_dump_debug_filepath(context)
    if debug_filepath:
        detailed_msg = f"See detailed message in {debug_filepath}."
    else:
        detailed_msg = "Enable verbose logs to see details."

    version = utils.get_version()
    public_schema = f"https://raw.githubusercontent.com/Netcracker/KubeMarine/{version}/kubemarine/{root_schema_resource}"
    hint = f"Inventory file{for_procedure} has incorrect format. {detailed_msg}\n" \
           f"To validate the file manually, you can use JSON schema {root_schema_uri}\n" \
           f"or its public alternative {public_schema}\n"

    msg = _error_msg(validator, errs[0])
    raise errors.FailException(msg, hint=hint)


def _resolve_errors(errs: List[jsonschema.ValidationError]) -> List[jsonschema.ValidationError]:
    key = _extended_relevance()
    for error in errs:
        _unnest_errors(error)

    errs.sort(key=key, reverse=True)
    outer_sorted = []
    for error in errs:
        outer_sorted.extend(_descend_errors(error))

    return outer_sorted


def _unnest_errors(error: jsonschema.ValidationError) -> None:
    if not error.context:
        return

    context: List[jsonschema.ValidationError] = error.context
    errors_by_subschema: Dict[Union[str, int], List[jsonschema.ValidationError]] = {}
    for child in context:
        _unnest_errors(child)
        errors_by_subschema.setdefault(child.schema_path[0], []).append(child)

    # jsonschema might show not very friendly messages for anyOf / oneOf.
    # In case of failed type or enum validation in each subschema, we can unnest the error to the higher level.
    # Note that possible remaining subschema errors will be discarded,
    # but it is acceptable for our current schemas implementation.
    subschemas_errors = list(errors_by_subschema.values())
    _unnest_type_subschema_errors(error, subschemas_errors)
    _unnest_enum_subschema_errors(error, subschemas_errors)
    _unnest_required_subschema_errors(error, subschemas_errors)


def _descend_errors(error: jsonschema.ValidationError) -> List[jsonschema.ValidationError]:
    if not error.context:
        return [error]

    # Here can be anyOf or oneOf with all not valid subschemas.
    # The idea is taken from jsonschema.exceptions.best_match,
    # with improved heuristic algorithm of choosing of the best subschema match.
    # 1. group errors by subschema
    # 2. find error with the max relevance in each group
    # 3. choose group with the min relevance of the error found in the previous step
    # 4. take the group instead of the parent error, sort by relevance reverse
    # 5. recurse if necessary

    context: List[jsonschema.ValidationError] = error.context
    errors_by_subschema: Dict[Union[str, int], List[jsonschema.ValidationError]] = {}
    for child in context:
        errors_by_subschema.setdefault(child.schema_path[0], []).append(child)

    key = _extended_relevance()

    for errors in errors_by_subschema.values():
        errors.sort(key=key, reverse=True)

    smallest_subschema_errors = sorted(errors_by_subschema.values(), key=lambda errors: key(errors[0]))

    # We don't recurse if two the smallest errors have the same relevance (i.e. if min == max == all).
    # This place is a candidate of future developing of better heuristic.
    if len(smallest_subschema_errors) >= 2 and key(smallest_subschema_errors[0][0]) == key(smallest_subschema_errors[1][0]):
        return [error]

    unnested_errors = []
    for err in smallest_subschema_errors[0]:
        unnested_errors.extend(_descend_errors(err))

    return unnested_errors


def _unnest_type_subschema_errors(error: jsonschema.ValidationError,
                                  subschemas_errors: List[List[jsonschema.ValidationError]]) -> None:
    if not error.context:
        return

    expected_types = OrderedSet[str]()
    for errs in subschemas_errors:
        for child in errs:
            if _validated_by(child, "type") and len(child.relative_path) == 0 and list(child.schema_path)[1:] == ["type"]:
                value = child.validator_value
                expected_types.update([value] if isinstance(value, str) else value)
                break
        else:  # not found error with "type" validation failed for root instance.
            break
    else:  # not found subschema not containing the necessary error, i.e. all subschemas have necessary error
        reprs = ", ".join(repr(type) for type in expected_types)
        for child in error.context:
            child.parent = None
        error.context = []
        error.validator = "type"  # type: ignore[assignment]
        error.validator_value = list(expected_types)
        error.schema_path[-1] = "type"
        error.message = f"{error.instance!r} is not of type {reprs}"
        subschemas_errors.clear()


def _unnest_enum_subschema_errors(error: jsonschema.ValidationError,
                                  subschemas_errors: List[List[jsonschema.ValidationError]]) -> None:
    if not error.context:
        return

    expected_elems = OrderedSet[str]()
    for errs in subschemas_errors:
        for child in errs:
            if _validated_by(child, "enum") and len(child.relative_path) == 0 and list(child.schema_path)[1:] == ["enum"]:
                expected_elems.update(child.validator_value)
                break
        else:  # not found error with "enum" validation failed for root instance.
            break
    else:  # not found subschema not containing the necessary error, i.e. all subschemas have necessary error
        for child in error.context:
            child.parent = None
        error.context = []
        error.validator = "enum"  # type: ignore[assignment]
        error.validator_value = list(expected_elems)
        error.schema_path[-1] = "enum"
        error.message = f"{error.instance!r} is not one of {list(expected_elems)!r}"
        subschemas_errors.clear()


def _unnest_required_subschema_errors(error: jsonschema.ValidationError,
                                      subschemas_errors: List[List[jsonschema.ValidationError]]) -> None:
    if not error.context:
        return

    required_sets: List[List[str]] = []
    for errs in subschemas_errors:
        if len(errs) > 1:
            # Support only the case when all subschema failed with the only 'required' validation.
            # This is a common pattern if "at least one property from the set is required" requirement
            # is moved to the separate "allOf (anyOf (required prop1, required prop2)) subschema.
            break
        for child in errs:
            if (_validated_by(child, "required") and len(child.relative_path) == 0
                    and list(child.schema_path)[1:] == ["required"] and len(child.validator_value) > 0
                    and child.validator_value != ['<<']):
                if isinstance(child.validator_value[0], str):
                    required_sets.append(child.validator_value)
                else:
                    required_sets.extend(child.validator_value)
                break
        else:  # not found error with "required" validation failed for root instance.
            break
    # not found subschema not containing the only necessary error,
    # i.e. all subschemas have the only necessary error
    else:
        initial_validator = cast(str, error.validator)

        for child in error.context:
            child.parent = None

        error.context = []
        error.validator = "required"  # type: ignore[assignment]
        # This breaks original typing.
        # Access to the validator_value should always be guarded with extra isinstance() checks.
        error.validator_value = required_sets
        error.schema_path[-1] = "required"
        error.message = (f"{'One' if initial_validator == 'oneOf' else 'At least one'} "
                         f"of the following property sets is required: {', '.join(map(str, required_sets))}")
        subschemas_errors.clear()


def _extended_relevance() -> Callable[[jsonschema.ValidationError], tuple]:
    # The extended relevance function is intended to improve heuristic for oneOf|anyOf,
    # when it is necessary to choose the most suitable branch.

    def relevance(error: jsonschema.ValidationError) -> tuple:
        relevance_value: tuple = jsonschema.exceptions.relevance(error)  # type: ignore[assignment]
        if error.parent is None:
            return relevance_value

        relevance_value = _apply_property_names_heuristic(error, relevance_value)
        relevance_value = _apply_list_merging_strong_heuristic(error, relevance_value)
        relevance_value = _apply_required_and_optional_properties_heuristic(error, relevance_value)

        return relevance_value

    return relevance


def _apply_property_names_heuristic(error: jsonschema.ValidationError, relevance_value: tuple) -> tuple:
    # jsonschema has type matching heuristic but it works bad for "propertyNames".
    # "propertyNames" does not introduce new path element and effectively verifies the object holding the property.
    # But an attempt to match the type happens for the "propertyNames" subschema.
    # The following heuristic resolves the simplest case of oneOf|anyOf(something, schema with "propertyNames" section)

    if len(error.relative_path) == 0 and "propertyNames" in error.schema_path:
        # "propertyNames" is validated only if the instance is "object".
        # See jsonschema._validators.propertyNames. So type is always matched.
        type_matched = True
        mutate_relevance = list(relevance_value)
        mutate_relevance[3] = not type_matched
        return tuple(mutate_relevance)

    return relevance_value


def _apply_list_merging_strong_heuristic(error: jsonschema.ValidationError, relevance_value: tuple) -> tuple:
    # Other conditions being equal, the error for list merging strategy has greater relevance,
    # because if the user specified '<<', we consider that he/she intends to use this advanced feature.

    is_list_merging = len(error.relative_path) == 0 and isinstance(error.schema, dict) \
                      and "properties" in error.schema and error.schema["properties"].keys() == {"<<"}

    return _append_relevance_element(relevance_value, is_list_merging)


def _apply_required_and_optional_properties_heuristic(error: jsonschema.ValidationError,
                                                      relevance_value: tuple) -> tuple:
    # Resolve oneOf|anyOf(object, object) ambiguity.
    # Currently only 'registry' section has such schema.
    # The chosen priority is:
    # 1. If all required properties are present, then the error has lower relevance.
    # 2. Otherwise, calculate proportion of matched properties.

    # By default, consider all required properties are present, and no properties are matched
    coeff: list = [False, 0]
    if len(error.relative_path) != 0 or not isinstance(error.schema, dict) or not isinstance(error.instance, dict):
        return _append_relevance_element(relevance_value, tuple(coeff))

    coeff[0] = not all(p in error.instance for p in error.schema.get("required", []))

    if len(error.schema.get("properties", {})) > 0:
        inverse_weight = len(error.schema["properties"])
        matched = set(error.instance.keys()).intersection(set(error.schema["properties"].keys()))
        # negative coefficient because more matched properties - more suitable subschema - less relevant error
        coeff[1] = - len(matched) / inverse_weight

    return _append_relevance_element(relevance_value, tuple(coeff))


def _append_relevance_element(relevance_value: tuple, elem: Hashable) -> tuple:
    mutate_relevance = list(relevance_value)
    mutate_relevance.append(elem)
    return tuple(mutate_relevance)


def _error_msg(validator: jsonschema.Draft7Validator, error: jsonschema.ValidationError) -> str:
    return dedent(
        f"""\
        {_friendly_msg(validator, error)}
        On inventory{utils.pretty_path(error.absolute_path)}
        """.rstrip()
    )


def _verbose_msg(validator: jsonschema.Draft7Validator, error: jsonschema.ValidationError) -> str:
    schema_path = utils.pretty_path(list(error.absolute_schema_path)[:-1])
    return dedent(
        f"""\
        {_friendly_msg(validator, error)}
        
        Failed validating {error.validator!r} in {schema_path}
        On inventory{utils.pretty_path(error.absolute_path)}
        """.rstrip()
    )


def _friendly_msg(validator: jsonschema.Draft7Validator, error: jsonschema.ValidationError) -> str:
    # pylint: disable=too-many-return-statements

    if _validated_by(error, 'minProperties'):
        return f"Number of properties is less than the minimum of {error.validator_value!r}. " \
               f"Property names: {list(error.instance)!r}"

    if _validated_by(error, 'maxProperties'):
        return f"Number of properties is greater than the maximum of {error.validator_value!r}. " \
               f"Property names: {list(error.instance)!r}"

    if _validated_by(error, 'minItems'):
        return f"Number of items equal to {len(error.instance)} is less than the minimum of {error.validator_value!r}"

    if _validated_by(error, 'maxItems'):
        return f"Number of items equal to {len(error.instance)} is greater than the maximum of {error.validator_value!r}"

    if _validated_by(error, 'type'):
        value = error.validator_value
        expected_types = [value] if isinstance(value, str) else value
        reprs = ", ".join(repr(type) for type in expected_types)
        try:
            for type_ in validator.TYPE_CHECKER._type_checkers:  # pylint: disable=protected-access
                if validator.is_type(error.instance, type_):  # type: ignore[no-untyped-call]
                    return f"Actual instance type is {type_!r}. Expected: {reprs}."
        except (AttributeError, TypeError, jsonschema.exceptions.UnknownType):
            # In current 3rd-party version the error should not appear, but let's still fall back to default behaviour
            pass

    if (_validated_by(error, 'not') and isinstance(error.validator_value, dict)
            and set(error.validator_value) == {'enum'} and "propertyNames" in error.schema_path):
        return f"Property name {error.instance!r} is unexpected"

    if _validated_by(error, 'enum'):
        if "propertyNames" not in error.schema_path:
            return f"Value should be one of {error.validator_value!r}"
        else:
            return f"Property name {error.instance!r} is not one of {error.validator_value!r}"

    return error.message


def _validated_by(error: jsonschema.ValidationError, expected: str) -> bool:
    return error.validator == expected  # type: ignore[comparison-overlap]



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import os
import re
from typing import Dict, List, Optional, Union

import ruamel.yaml
import yaml
from jinja2 import Template

from kubemarine.core import utils
from kubemarine.core.cluster import KubernetesCluster, EnrichmentStage, enrichment
from kubemarine.core.group import NodeGroup, RunnersGroupResult
from kubemarine.core.yaml_merger import default_merger
from kubemarine.kubernetes import components
from kubemarine.plugins import builtin

privileged_policy_filename = "privileged.yaml"
policies_file_path = "./resources/psp/"

admission_template = "./templates/admission.yaml.j2"
admission_dir = "/etc/kubernetes/pki"
admission_path = "%s/admission.yaml" % admission_dir

psp_list_option = "psp-list"
roles_list_option = "roles-list"
bindings_list_option = "bindings-list"

provided_oob_policies = ["default", "host-network", "anyuid"]

valid_modes = ['enforce', 'audit', 'warn']
valid_versions_templ = r"^v1\.\d{1,2}$"

ERROR_INCONSISTENT_INVENTORIES = "Procedure config and cluster config are inconsistent. Please check 'admission' option"
ERROR_CHANGE_OOB_PSP_DISABLED = "OOB policies can not be configured when security is disabled"
ERROR_PSS_BOTH_STATES_DISABLED = ("both 'pod-security' in procedure config and current config are 'disabled'. "
                                  "There is nothing to change")

# TODO: When KubeMarine is not support Kubernetes version lower than 1.25, the PSP implementation code should be deleted 


def support_pss_only(cluster: KubernetesCluster) -> bool:
    return components.kubernetes_minor_release_at_least(cluster.inventory, "v1.25")


def is_pod_security_unconditional(cluster: KubernetesCluster) -> bool:
    return components.kubernetes_minor_release_at_least(cluster.inventory, "v1.28")


@enrichment(EnrichmentStage.FULL)
def enrich_inventory_psp(cluster: KubernetesCluster) -> None:
    inventory = cluster.inventory

    # validate custom
    custom_policies = inventory["rbac"]["psp"]["custom-policies"]
    verify_custom(custom_policies)

    # do not perform enrichment if security disabled
    if not is_security_enabled(inventory):
        return

    # if security enabled, then add PodSecurityPolicy admission plugin
    enabled_admissions = inventory["services"]["kubeadm"]["apiServer"]["extraArgs"]["enable-admission-plugins"]
    if 'PodSecurityPolicy' not in enabled_admissions:
        enabled_admissions = "%s,PodSecurityPolicy" % enabled_admissions
        inventory["services"]["kubeadm"]["apiServer"]["extraArgs"]["enable-admission-plugins"] = enabled_admissions


@enrichment(EnrichmentStage.FULL)
def enrich_inventory_pss(cluster: KubernetesCluster) -> None:
    inventory = cluster.inventory
    # check flags, enforce and logs parameters
    kubernetes_version = inventory["services"]["kubeadm"]["kubernetesVersion"]

    for item, state in cluster.inventory["rbac"]["pss"]["defaults"].items():
        if item.endswith("version"):
            verify_version(item, state, kubernetes_version)

    if not is_security_enabled(inventory):
        return

    # add extraArgs to kube-apiserver config
    extra_args = inventory["services"]["kubeadm"]["apiServer"]["extraArgs"]
    if not is_pod_security_unconditional(cluster):
        enabled_admissions = extra_args.get("feature-gates")
        if enabled_admissions:
            if 'PodSecurity=true' not in enabled_admissions:
                enabled_admissions = "%s,PodSecurity=true" % enabled_admissions
        else:
            enabled_admissions = "PodSecurity=true"

        extra_args["feature-gates"] = enabled_admissions

    extra_args["admission-control-config-file"] = admission_path


@enrichment(EnrichmentStage.FULL)
def enrich_inventory(cluster: KubernetesCluster) -> None:
    default_admission = "pss" if support_pss_only(cluster) else "psp"
    admission_impl = cluster.inventory["rbac"].setdefault("admission", default_admission)

    if support_pss_only(cluster) and (
            admission_impl == "psp"
            or cluster.previous_inventory["rbac"]["admission"] == "psp"
    ):
        raise Exception("PSP is not supported in Kubernetes v1.25 or higher")

    if admission_impl == "psp":
        del cluster.inventory["rbac"]["pss"]
        enrich_inventory_psp(cluster)
    elif admission_impl == "pss":
        del cluster.inventory["rbac"]["psp"]
        enrich_inventory_pss(cluster)


@enrichment(EnrichmentStage.PROCEDURE, procedures=['manage_psp'])
def verify_psp_enrichment(cluster: KubernetesCluster) -> None:
    procedure_config = cluster.procedure_inventory["psp"]

    # validate added custom
    custom_add_policies = procedure_config.get("add-policies", {})
    verify_custom(custom_add_policies)

    # validate deleted custom
    custom_delete_policies = procedure_config.get("delete-policies", {})
    verify_custom(custom_delete_policies)

    # forbid managing OOB if security will be disabled
    previous_psp = cluster.previous_inventory["rbac"]["psp"]
    psp = cluster.inventory["rbac"]["psp"]
    if psp["pod-security"] == "disabled" and previous_psp["oob-policies"] != psp["oob-policies"]:
        raise Exception(ERROR_CHANGE_OOB_PSP_DISABLED)


def verify_custom(custom_scope: Dict[str, List[dict]]) -> None:
    psp_list = custom_scope.get(psp_list_option, None)
    if psp_list:
        verify_custom_list(psp_list, "PSP")

    roles_list = custom_scope.get(roles_list_option, None)
    if roles_list:
        verify_custom_list(roles_list, "role")

    bindings_list = custom_scope.get(bindings_list_option, None)
    if bindings_list:
        verify_custom_list(bindings_list, "binding")


def verify_custom_list(custom_list: List[dict], type_: str) -> None:
    for item in custom_list:
        # forbid using 'oob-' prefix in order to avoid conflicts of our policies and users policies
        if item["metadata"]["name"].startswith("oob-"):
            raise Exception("Name %s is not allowed for custom %s" % (item["metadata"]["name"], type_))


def verify_version(owner: str, version: str, kubernetes_version: str) -> None:
    # check Kubernetes version and admission config matching
    if version != "latest":
        result = re.match(valid_versions_templ, version)
        if result is None:
            raise Exception("incorrect Kubernetes version %s, valid version(for example): v1.23" % owner)
        if utils.minor_version_key(version) > utils.version_key(kubernetes_version)[0:2]:
            raise Exception("%s version must not be higher than Kubernetes version" % owner)


@enrichment(EnrichmentStage.PROCEDURE, procedures=['manage_psp'])
def manage_psp_enrichment(cluster: KubernetesCluster) -> None:
    procedure_config = cluster.procedure_inventory["psp"]

    current_config = cluster.inventory.setdefault("rbac", {}).setdefault("psp", {})

    # Perform custom-policies lists changes.
    # Perform changes only if there are any "custom-policies" or "add-policies" in inventory,
    # do not perform changes if only "delete-policies" defined, there is nothing to delete from inventory in this case.
    adding_custom_policies = procedure_config.get("add-policies", {})
    deleting_custom_policies = procedure_config.get("delete-policies", {})
    existing_custom_policies = current_config.get("custom-policies", {})
    if existing_custom_policies or adding_custom_policies:
        # if custom policies are not defined in inventory, then we need to create custom policies ourselves
        current_config["custom-policies"] = merge_custom_policies(
            existing_custom_policies,
            utils.deepcopy_yaml(adding_custom_policies),
            utils.deepcopy_yaml(deleting_custom_policies))

    # merge flags from procedure config and cluster config
    current_config["pod-security"] = procedure_config.get("pod-security", current_config.get("pod-security", "enabled"))
    if "oob-policies" in procedure_config:
        default_merger.merge(current_config.setdefault("oob-policies", {}), procedure_config["oob-policies"])


def merge_custom_policies(old_policies: Dict[str, List[dict]],
                          added_policies: Dict[str, List[dict]],
                          deleted_policies: Dict[str, List[dict]]) -> Dict[str, List[dict]]:
    for option in (psp_list_option, roles_list_option, bindings_list_option):
        old_policies[option] = merge_policy_lists(old_policies.get(option, []),
                                                  added_policies.get(option, []),
                                                  deleted_policies.get(option, []))

    return old_policies


def merge_policy_lists(old_list: List[dict], added_list: List[dict], deleted_list: List[dict]) -> List[dict]:
    resulting_list = added_list
    added_names_list = [item["metadata"]["name"] for item in added_list]
    deleted_names_list = [item["metadata"]["name"] for item in deleted_list]
    for old_item in old_list:
        old_item_name = old_item["metadata"]["name"]
        if old_item_name in added_names_list or old_item_name in deleted_names_list:
            # skip old item, since it was either deleted, replaced by new item, or deleted and then replaced
            continue
        # old item is nor deleted, nor updated, then we need to preserve it in resulting list
        resulting_list.append(old_item)

    return resulting_list


def install_psp_task(cluster: KubernetesCluster) -> None:
    if not is_security_enabled(cluster.inventory):
        cluster.log.debug("Pod security disabled, skipping policies installation...")
        return

    first_control_plane = cluster.nodes["control-plane"].get_first_member()

    cluster.log.debug("Installing OOB policies...")
    first_control_plane.call(manage_policies,
                      manage_type="apply",
                      manage_scope=resolve_oob_scope(cluster.inventory["rbac"]["psp"]["oob-policies"], "enabled"))

    cluster.log.debug("Installing custom policies...")
    first_control_plane.call(manage_policies,
                      manage_type="apply",
                      manage_scope=cluster.inventory["rbac"]["psp"]["custom-policies"])


def delete_custom_task(cluster: KubernetesCluster) -> None:
    if "delete-policies" not in cluster.procedure_inventory["psp"]:
        cluster.log.debug("No 'delete-policies' specified, skipping...")
        return

    cluster.log.debug("Deleting custom 'delete-policies'")
    first_control_plane = cluster.nodes["control-plane"].get_first_member()
    first_control_plane.call(manage_policies,
                      manage_type="delete",
                      manage_scope=cluster.procedure_inventory["psp"]["delete-policies"])


def add_custom_task(cluster: KubernetesCluster) -> None:
    if "add-policies" not in cluster.procedure_inventory["psp"]:
        cluster.log.debug("No 'add-policies' specified, skipping...")
        return

    cluster.log.debug("Applying custom 'add-policies'")
    first_control_plane = cluster.nodes["control-plane"].get_first_member()
    first_control_plane.call(manage_policies,
                      manage_type="apply",
                      manage_scope=cluster.procedure_inventory["psp"]["add-policies"])


def reconfigure_psp_task(cluster: KubernetesCluster) -> None:
    if is_security_enabled(cluster.inventory):
        reconfigure_oob_policies(cluster)
        reconfigure_plugin(cluster)
    else:
        # If target state is disabled, firstly remove PodSecurityPolicy admission plugin thus disabling the security.
        # It is necessary because while security on one control-plane is disabled,
        # it is still enabled on the other control-planes, and the policies should still work.
        reconfigure_plugin(cluster)
        reconfigure_oob_policies(cluster)


def reconfigure_oob_policies(cluster: KubernetesCluster) -> None:
    previous_psp = cluster.previous_inventory["rbac"]["psp"]
    psp = cluster.inventory["rbac"]["psp"]
    target_state = psp["pod-security"]
    target_config = psp["oob-policies"]

    # reconfigure OOB only if state will be changed, or OOB configuration was changed
    if previous_psp["pod-security"] == target_state and previous_psp["oob-policies"] == target_config:
        cluster.log.debug("No need to reconfigure OOB policies, skipping...")
        return

    first_control_plane = cluster.nodes["control-plane"].get_first_member()

    cluster.log.debug("Deleting all OOB policies...")
    first_control_plane.call(delete_privileged_policy)
    first_control_plane.call(manage_policies,
                             manage_type="delete",
                             manage_scope=resolve_oob_scope(target_config, "all"))

    if target_state == "disabled":
        cluster.log.debug("Security disabled, OOB will not be recreated")
        return

    cluster.log.debug("Recreating all OOB policies...")
    first_control_plane.call(apply_privileged_policy)
    first_control_plane.call(manage_policies,
                             manage_type="apply",
                             manage_scope=resolve_oob_scope(target_config, "enabled"))


def reconfigure_plugin(cluster: KubernetesCluster) -> None:
    previous_psp = cluster.previous_inventory["rbac"]["psp"]
    psp = cluster.inventory["rbac"]["psp"]

    if previous_psp["pod-security"] == psp["pod-security"]:
        cluster.log.debug("Security plugin will not be reconfigured")
        return

    cluster.nodes['control-plane'].call(components.reconfigure_components, components=['kube-apiserver'])


def restart_pods_task(cluster: KubernetesCluster) -> None:
    from kubemarine import kubernetes  # pylint: disable=cyclic-import

    if cluster.context.get('initial_procedure') == 'manage_pss':
        # check if pods restart is enabled
        is_restart = cluster.procedure_inventory.get("restart-pods", False)
        if not is_restart:
            cluster.log.debug("'restart-pods' is disabled, pods won't be restarted")
            return

    first_control_plane = cluster.nodes["control-plane"].get_first_member()

    cluster.log.debug("Drain-Uncordon all nodes to restart pods")
    kube_nodes = cluster.nodes["control-plane"].include_group(cluster.nodes["worker"])
    for node in kube_nodes.get_ordered_members_list():
        first_control_plane.sudo(
            kubernetes.prepare_drain_command(cluster, node.get_node_name(), disable_eviction=False),
            hide=False)
        first_control_plane.sudo("kubectl uncordon %s" % node.get_node_name(), hide=False)

    cluster.log.debug("Restarting daemon-sets...")
    daemon_sets = ruamel.yaml.YAML().load(list(first_control_plane.sudo("kubectl get ds -A -o yaml").values())[0].stdout)
    for ds in daemon_sets["items"]:
        first_control_plane.sudo("kubectl rollout restart ds %s -n %s" % (ds["metadata"]["name"], ds["metadata"]["namespace"]))

    # we do not know to wait for, only for system pods maybe
    kube_nodes.call(components.wait_for_pods)


def is_security_enabled(inventory: dict) -> bool:
    admission_impl = inventory['rbac']['admission']
    target_state = "disabled"
    if admission_impl == "psp":
        target_state = inventory["rbac"]["psp"]["pod-security"]
    elif admission_impl == "pss":
        target_state = inventory["rbac"]["pss"]["pod-security"]

    return target_state == "enabled"


def apply_privileged_policy(group: NodeGroup) -> RunnersGroupResult:
    return manage_privileged_from_file(group, privileged_policy_filename, "apply")


def delete_privileged_policy(group: NodeGroup) -> RunnersGroupResult:
    return manage_privileged_from_file(group, privileged_policy_filename, "delete")


def apply_admission(group: NodeGroup) -> None:
    admission_impl = group.cluster.inventory['rbac']['admission']
    if admission_impl == "psp" and is_security_enabled(group.cluster.inventory):
        group.cluster.log.debug("Setting up privileged psp...")
        apply_privileged_policy(group)


def manage_privileged_from_file(group: NodeGroup, filename: str, manage_type: str) -> RunnersGroupResult:
    if manage_type not in ["apply", "delete"]:
        raise Exception("unexpected manage type for privileged policy")
    privileged_policy = utils.read_internal(os.path.join(policies_file_path, filename))
    remote_path = utils.get_remote_tmp_path(filename)
    group.put(io.StringIO(privileged_policy), remote_path, backup=True, sudo=True)

    return group.sudo("kubectl %s -f %s" % (manage_type, remote_path), warn=True)


def resolve_oob_scope(oob_policies_conf: Dict[str, str], selector: str) -> Dict[str, List[dict]]:
    result: Dict[str, List[dict]] = {
        psp_list_option: [],
        roles_list_option: [],
        bindings_list_option: []
    }

    loaded_oob_policies = load_oob_policies_files()

    for key, value in oob_policies_conf.items():
        if selector in (value, "all"):
            policy = loaded_oob_policies[key]
            if "psp" in policy:
                result[psp_list_option].append(policy["psp"])
            if "role" in policy:
                result[roles_list_option].append(policy["role"])
            if "binding" in policy:
                result[bindings_list_option].append(policy["binding"])

    return result


def load_oob_policies_files() -> Dict[str, dict]:
    oob_policies = {}
    for oob_name in provided_oob_policies:
        local_path = os.path.join(policies_file_path, "%s.yaml" % oob_name)
        with utils.open_internal(local_path) as stream:
            oob_policies[oob_name] = yaml.safe_load(stream)

    return oob_policies


def manage_policies(group: NodeGroup, manage_type: str,
                    manage_scope: Dict[str, List[dict]]) -> Optional[RunnersGroupResult]:
    psp_to_manage = manage_scope.get(psp_list_option, None)
    roles_to_manage = manage_scope.get(roles_list_option, None)
    bindings_to_manage = manage_scope.get(bindings_list_option, None)

    if not psp_to_manage and not roles_to_manage and not bindings_to_manage:
        group.cluster.log.verbose("No policies to %s" % manage_type)
        return None

    template = collect_policies_template(psp_to_manage, roles_to_manage, bindings_to_manage)
    remote_path = utils.get_remote_tmp_path()
    group.put(io.StringIO(template), remote_path, backup=True, sudo=True)
    result = group.sudo("kubectl %s -f %s" % (manage_type, remote_path), warn=True)
    group.sudo("rm -f %s" % remote_path)
    return result


def collect_policies_template(psp_list: Optional[List[dict]],
                              roles_list: Optional[List[dict]],
                              bindings_list: Optional[List[dict]]) -> str:
    buf = io.StringIO()
    if psp_list:
        for psp in psp_list:
            yaml.dump(psp, buf)
            buf.write("\n---\n")
    if roles_list:
        for role in roles_list:
            yaml.dump(role, buf)
            buf.write("\n---\n")
    if bindings_list:
        for binding in bindings_list:
            yaml.dump(binding, buf)
            buf.write("\n---\n")
    return buf.getvalue()


def install(cluster: KubernetesCluster) -> None:
    admission_impl = cluster.inventory['rbac']['admission']
    if admission_impl == "psp":
        install_psp_task(cluster)


@enrichment(EnrichmentStage.PROCEDURE, procedures=['manage_pss'])
def verify_pss_enrichment(cluster: KubernetesCluster) -> None:
    procedure_config = cluster.procedure_inventory["pss"]
    kubernetes_version = cluster.previous_inventory["services"]["kubeadm"]["kubernetesVersion"]

    if (cluster.previous_inventory["rbac"]["pss"]["pod-security"] == "disabled"
            and cluster.inventory["rbac"]["pss"]["pod-security"] == "disabled"):
        raise Exception(ERROR_PSS_BOTH_STATES_DISABLED)

    if "namespaces" in procedure_config:
        for namespace_item in procedure_config["namespaces"]:
            # check if the namespace has its own profiles
            if isinstance(namespace_item, dict):
                profiles = list(namespace_item.values())[0]
                for item in profiles:
                    if item.endswith("version"):
                        verify_version(item, profiles[item], kubernetes_version)
    if "namespaces_defaults" in procedure_config:
        for item in procedure_config["namespaces_defaults"]:
            if item.endswith("version"):
                verify_version(item, procedure_config["namespaces_defaults"][item], kubernetes_version)


@enrichment(EnrichmentStage.PROCEDURE, procedures=['manage_psp', 'manage_pss'])
def manage_enrichment(cluster: KubernetesCluster) -> None:
    check_inventory(cluster)
    admission_impl = cluster.previous_inventory['rbac']['admission']
    if admission_impl == "psp":
        manage_psp_enrichment(cluster)
    elif admission_impl == "pss":
        manage_pss_enrichment(cluster)


@enrichment(EnrichmentStage.PROCEDURE, procedures=['manage_psp', 'manage_pss'])
def verify_manage_enrichment(cluster: KubernetesCluster) -> None:
    admission_impl = cluster.previous_inventory['rbac']['admission']
    if admission_impl == "psp":
        verify_psp_enrichment(cluster)
    elif admission_impl == "pss":
        verify_pss_enrichment(cluster)


def manage_pss(cluster: KubernetesCluster) -> None:
    control_planes = cluster.nodes["control-plane"]

    target_state = cluster.inventory["rbac"]["pss"]["pod-security"]

    # set labels for predefined plugins namespaces and namespaces defined in procedure config
    label_namespace_pss(cluster)

    # copy admission config on control-planes
    copy_pss(control_planes)

    # Admission configuration may change.
    # Force kube-apiserver pod restart, then wait for API server to become available.
    force_restart = True

    # Extra args of API may change, need to reconfigure the API server.
    # See enrich_inventory_pss()
    control_planes.call(components.reconfigure_components,
                        components=['kube-apiserver'], force_restart=force_restart)

    if target_state == 'disabled':
        # erase PSS admission config
        cluster.log.debug("Erase admission configuration... %s" % admission_path)
        control_planes.sudo("rm -f %s" % admission_path, warn=True)


@enrichment(EnrichmentStage.PROCEDURE, procedures=['manage_pss'])
def manage_pss_enrichment(cluster: KubernetesCluster) -> None:
    procedure_config = cluster.procedure_inventory["pss"]

    current_config = cluster.inventory.setdefault("rbac", {}).setdefault("pss", {})

    # merge flags from procedure config and cluster config
    current_config["pod-security"] = procedure_config["pod-security"]
    if "defaults" in procedure_config:
        default_merger.merge(current_config.setdefault("defaults", {}), procedure_config["defaults"])
    if "exemptions" in procedure_config:
        default_merger.merge(current_config.setdefault("exemptions", {}), procedure_config["exemptions"])


def generate_pss(cluster: KubernetesCluster) -> str:
    defaults = cluster.inventory["rbac"]["pss"]["defaults"]
    exemptions = cluster.inventory["rbac"]["pss"]["exemptions"]
    return Template(utils.read_internal(admission_template))\
        .render(defaults=defaults, exemptions=exemptions)


def copy_pss(group: NodeGroup) -> Optional[RunnersGroupResult]:
    if group.cluster.inventory['rbac']['admission'] != "pss":
        return None

    if not is_security_enabled(group.cluster.inventory):
        group.cluster.log.debug("Pod security disabled, skipping pod admission installation...")
        return None

    # create admission config from template and cluster.yaml
    admission_config = generate_pss(group.cluster)

    # put admission config on every control-planes
    group.cluster.log.debug(f"Copy admission config to {admission_path}")
    group.put(io.StringIO(admission_config), admission_path, backup=True, sudo=True, mkdir=True)

    return group.sudo(f'ls -la {admission_path}')


def _get_default_labels(profile: str) -> Dict[str, str]:
    return {f"pod-security.kubernetes.io/{k}": v
            for mode in valid_modes
            for k, v in ((mode, profile), (f'{mode}-version', 'latest'))}


def get_labels_to_ensure_profile(inventory: dict, profile: str) -> Dict[str, str]:
    enforce_profile: str = inventory['rbac']['pss']['defaults']['enforce']
    if (enforce_profile == 'restricted' and profile != 'restricted'
            or enforce_profile == 'baseline' and profile == 'privileged'):
        return _get_default_labels(profile)

    return {}


def label_namespace_pss(cluster: KubernetesCluster) -> None:
    security_enabled = is_security_enabled(cluster.inventory)
    first_control_plane = cluster.nodes["control-plane"].get_first_member()
    # set/delete labels on predifined plugins namsespaces
    for ns_name, profile in builtin.get_namespace_to_necessary_pss_profiles(cluster).items():
        target_labels = get_labels_to_ensure_profile(cluster.inventory, profile)
        if security_enabled and target_labels:
            cluster.log.debug(f"Set PSS labels for profile {profile} on namespace {ns_name}")
            command = "kubectl label ns {namespace} {lk}={lv} --overwrite"

        else:  # Pod Security is disabled or default labels are not necessary
            cluster.log.debug(f"Delete PSS labels from namespace {ns_name}")
            command = "kubectl label ns {namespace} {lk}- || true"
            target_labels = _get_default_labels(profile)

        for lk, lv in target_labels.items():
            first_control_plane.sudo(command.format(namespace=ns_name, lk=lk, lv=lv))

    procedure_config = cluster.procedure_inventory["pss"]
    namespaces: List[Union[str, Dict[str, dict]]] = procedure_config.get("namespaces")
    # get the list of namespaces that should be labeled then set/delete labels
    if namespaces:
        default_modes = {}
        # check if procedure config has default values for labels
        namespaces_defaults = procedure_config.get("namespaces_defaults")
        if namespaces_defaults:
            for default_mode in namespaces_defaults:
                default_modes[default_mode] = namespaces_defaults[default_mode]
        for namespace in namespaces:
            # define name of namespace
            if isinstance(namespace, dict):
                ns_name = list(namespace.keys())[0]
            else:
                ns_name = namespace
            if security_enabled:
                if default_modes:
                    # set labels that are set in default section
                    cluster.log.debug(f"Set PSS labels on {ns_name} namespace from defaults")
                    for mode, value in default_modes.items():
                        first_control_plane.sudo(f"kubectl label ns {ns_name} "
                                f"pod-security.kubernetes.io/{mode}={value} --overwrite")
                if isinstance(namespace, dict):
                    # set labels that are set in namespaces section
                    cluster.log.debug(f"Set PSS labels on {ns_name} namespace")
                    for item in list(namespace[ns_name]):
                        first_control_plane.sudo(f"kubectl label ns {ns_name} " 
                                    f"pod-security.kubernetes.io/{item}={namespace[ns_name][item]} --overwrite")
            else:
                # delete labels that are set in default section
                if default_modes:
                    cluster.log.debug(f"Delete PSS labels on {ns_name} namespace from defaults")
                    for mode in default_modes:
                        first_control_plane.sudo(f"kubectl label ns {ns_name} pod-security.kubernetes.io/{mode}-")
                # delete labels that are set in namespaces section
                cluster.log.debug(f"Delete PSS labels on {ns_name} namespace")
                if isinstance(namespace, dict):
                    for item in list(namespace[ns_name]):
                        first_control_plane.sudo(f"kubectl label ns {ns_name} "
                                    f"pod-security.kubernetes.io/{item}-")


def check_inventory(cluster: KubernetesCluster) -> None:
    # check if 'admission' option in cluster.yaml and procedure.yaml are inconsistent
    previous_admission = cluster.previous_inventory["rbac"]["admission"]
    procedure = cluster.context['initial_procedure']
    if (procedure == 'manage_pss' and previous_admission != "pss"
            or procedure == 'manage_psp' and previous_admission != "psp"):
        raise Exception(ERROR_INCONSISTENT_INVENTORIES)



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from typing import Dict, List

import fabric  # type: ignore[import-untyped]

from kubemarine.core import static

Connections = Dict[str, fabric.connection.Connection]


class ConnectionPool:
    def __init__(self, nodes: Dict[str, dict], gateway_nodes: Dict[str, dict], hosts: List[str]):
        self._nodes = nodes
        self._gateway_nodes = gateway_nodes
        self._connections: Connections = {ip: self._create_connection(ip) for ip in hosts}

    def get_node(self, ip: str) -> dict:
        node = self._nodes.get(ip)
        if node is None:
            raise Exception("Failed to find suitable node to connect to by address %s" % ip)

        return node

    def get_connection(self, ip: str) -> fabric.connection.Connection:
        conn = self._connections.get(ip)
        if conn is None:
            raise Exception(f'Connection for {ip} is not registered')

        return conn

    def close(self) -> None:
        for conn in self._connections.values():
            conn.close()

        self._connections.clear()

    def _create_connection_from_details(self, ip: str, conn_details: dict,
                                        gateway: fabric.connection.Connection = None,
                                        inline_ssh_env: bool = True) -> fabric.connection.Connection:

        creds={}
        if conn_details.get('keyfile'):
            creds['key_filename'] = os.path.expanduser(conn_details['keyfile'])
        elif conn_details.get('password'):
            creds['password'] = conn_details.get('password')
        cfg = fabric.Config(overrides={'run': {'encoding': "utf-8"}})
        return fabric.connection.Connection(
            host=ip,
            user=conn_details.get('username', static.GLOBALS['connection']['defaults']['username']),
            gateway=gateway,
            port=conn_details.get('connection_port', static.GLOBALS['connection']['defaults']['port']),
            config=cfg,
            connect_timeout=conn_details.get('connection_timeout',
                                             static.GLOBALS['connection']['defaults']['timeout']),
            connect_kwargs=creds,
            inline_ssh_env=inline_ssh_env
        )

    def _create_connection(self, ip: str) -> fabric.connection.Connection:
        node = self.get_node(ip)

        if node.get('keyfile') is None and node.get('password') is None:
            raise Exception('There is neither keyfile nor password specified in configfile for node \'%s\'' % node['name'])

        gateway = None
        if 'gateway' in node:
            gateway = self._get_gateway_node_connection(node['gateway'])

        return self._create_connection_from_details(ip, node, gateway=gateway)

    def _get_gateway_node_connection(self, name: str) -> fabric.connection.Connection:
        # Create new connection instance each time even if it is the same gateway node.
        # This is necessary to not share the same gateway connection instance in multiple threads

        gateway = self._gateway_nodes.get(name)
        if gateway is None:
            raise Exception('Requested gateway \'%s\' is not found in configfile' % name)

        if gateway.get('address') is None:
            raise Exception('There is no address specified in configfile for gateway \'%s\'' % name)
        if gateway.get('keyfile') is None:
            raise Exception('There is no keyfile specified in configfile for gateway \'%s\'' % name)

        # todo since we have no workaround for gateway connections currently,
        #  probably we need different default connection timeout
        return self._create_connection_from_details(gateway["address"], gateway, inline_ssh_env=False)



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from kubemarine.core.cluster import KubernetesCluster, EnrichmentStage, enrichment


@enrichment(EnrichmentStage.ALL)
def controlplane_node_enrichment(cluster: KubernetesCluster) -> None:
    """
    Enriched inventory should have both the 'master' and the 'control-plane' roles for backward compatibility
    The 'control-plane' role is used instead of 'master' role since Kubernetes v1.24
    """
    for node in cluster.inventory["nodes"]:
        node_info = node['name']
        if "master" in node["roles"] and "control-plane" not in node["roles"]:
            cluster.log.debug(f"The 'control-plane' role will be added for {node_info}")
            cluster.log.warning(f"Node {node_info} has legacy role 'master'. "
                                f"Please use 'control-plane' instead.")
            node["roles"].append("control-plane")
        if "control-plane" in node["roles"] and "master" not in node["roles"]:
            cluster.log.debug(f"The 'master' role will be added for {node_info}")
            node["roles"].append("master")


def controlplane_finalize_inventory(cluster: KubernetesCluster, inventory: dict) -> dict:
    """
    Delete 'control-plane' and 'master' roles before inventory saving if they are not set in 'cluster.yaml'
    """
    # remove 'master' role from inventory before dump
    for node in inventory["nodes"]:
        if 'master' in node["roles"]:
            node_info = node['name']
            cluster.log.verbose(f"The 'master' role will be removed for {node_info} node before saving")
            node["roles"].remove("master")

    return inventory



# Copyright 2021-2023 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from textwrap import dedent

import yaml

from kubemarine import packages
from kubemarine.core import yaml_merger, errors
from kubemarine.core.action import Action
from kubemarine.core.cluster import EnrichmentStage
from kubemarine.core.patch import InventoryOnlyPatch
from kubemarine.core.resources import DynamicResources

OLD_SYSCTL_DEFAULTS = """
    net.bridge.bridge-nf-call-iptables: 1
    net.ipv4.ip_forward: 1
    net.ipv4.ip_nonlocal_bind: 1
    net.ipv4.conf.all.route_localnet: 1
    net.bridge.bridge-nf-call-ip6tables: '{% if not nodes[0]["internal_address"]|isipv4 %}1{% endif %}'
    net.ipv6.conf.all.forwarding: '{% if not nodes[0]["internal_address"]|isipv4 %}1{% endif %}'
    net.ipv6.ip_nonlocal_bind: '{% if not nodes[0]["internal_address"]|isipv4 %}1{% endif %}'
    net.netfilter.nf_conntrack_max: 1000000
    kernel.panic: 10
    vm.overcommit_memory: 1
    kernel.panic_on_oops: 1
"""

OLD_MODPROBE_DEFAULTS = """
    rhel:
      - br_netfilter
      - '{% if not nodes[0]["internal_address"]|isipv4 %}nf_conntrack_ipv6{% else %}nf_conntrack{% endif %}'
      - '{% if not nodes[0]["internal_address"]|isipv4 %}ip6table_filter{% endif %}'
      - '{% if not nodes[0]["internal_address"]|isipv4 %}nf_nat_masquerade_ipv6{% endif %}'
      - '{% if not nodes[0]["internal_address"]|isipv4 %}nf_reject_ipv6{% endif %}'
      - '{% if not nodes[0]["internal_address"]|isipv4 %}nf_defrag_ipv6{% endif %}'
    rhel8:
      - br_netfilter
      - nf_conntrack
      - '{% if not nodes[0]["internal_address"]|isipv4 %}ip6table_filter{% endif %}'
      - '{% if not nodes[0]["internal_address"]|isipv4 %}nf_nat{% endif %}'
      - '{% if not nodes[0]["internal_address"]|isipv4 %}nf_reject_ipv6{% endif %}'
      - '{% if not nodes[0]["internal_address"]|isipv4 %}nf_defrag_ipv6{% endif %}'
    rhel9:
      - br_netfilter
      - nf_conntrack
      - '{% if not nodes[0]["internal_address"]|isipv4 %}ip6table_filter{% endif %}'
      - '{% if not nodes[0]["internal_address"]|isipv4 %}nf_nat{% endif %}'
      - '{% if not nodes[0]["internal_address"]|isipv4 %}nf_reject_ipv6{% endif %}'
      - '{% if not nodes[0]["internal_address"]|isipv4 %}nf_defrag_ipv6{% endif %}'
    debian:
      - br_netfilter
      - nf_conntrack
      - '{% if not nodes[0]["internal_address"]|isipv4 %}ip6table_filter{% endif %}'
      - '{% if not nodes[0]["internal_address"]|isipv4 %}nf_nat{% endif %}'
      - '{% if not nodes[0]["internal_address"]|isipv4 %}nf_reject_ipv6{% endif %}'
      - '{% if not nodes[0]["internal_address"]|isipv4 %}nf_defrag_ipv6{% endif %}'
"""


class TheAction(Action):
    def __init__(self) -> None:
        super().__init__("Set previous default kernel parameters and modules", recreate_inventory=True)

    def run(self, res: DynamicResources) -> None:
        inventory = res.inventory()
        os_family = res.cluster(EnrichmentStage.LIGHT).get_os_family()
        if os_family not in packages.get_associations_os_family_keys():
            raise errors.KME("KME0012", procedure='migrate_kubemarine')

        sysctl_defaults = yaml.safe_load(OLD_SYSCTL_DEFAULTS)
        sysctl_config = inventory.setdefault('services', {}).setdefault('sysctl', {})
        for k, v in sysctl_defaults.items():
            if k not in sysctl_config:
                sysctl_config[k] = v

        default_list = yaml.safe_load(OLD_MODPROBE_DEFAULTS)[os_family]
        modprobe_config = inventory.setdefault('services', {}).setdefault('modprobe', {})
        if os_family not in modprobe_config:
            modprobe_config[os_family] = default_list
        else:
            modules_list = modprobe_config[os_family]
            strategy, pos = yaml_merger.get_strategy_position(modules_list, ['services', 'modprobe', os_family])
            if strategy == 'merge':
                modules_list[pos:(pos+1)] = default_list


class PreserveCompatibilityKernel(InventoryOnlyPatch):
    def __init__(self) -> None:
        super().__init__("preserve_compatibility_kernel")

    @property
    def action(self) -> Action:
        return TheAction()

    @property
    def description(self) -> str:
        return dedent(
            f"""\
            Set previous default kernel parameters and modules in the user inventory.
            
            This keeps as-is all parameters and modules previously installed by Kubemarine.
            This also preserves dynamic references to the default kernel parameters in jinja templates.
            """.rstrip()
        )



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Optional

from kubemarine.core import log
from kubemarine.core.cluster import KubernetesCluster, EnrichmentStage, enrichment
from kubemarine.plugins import calico, manifest
from kubemarine.plugins.calico import CalicoManifestProcessor, CalicoApiServerManifestProcessor
from kubemarine.plugins.kubernetes_dashboard import get_dashboard_manifest_processor
from kubemarine.plugins.local_path_provisioner import LocalPathProvisionerManifestProcessor
from kubemarine.plugins.manifest import Identity
from kubemarine.plugins.nginx_ingress import get_ingress_nginx_manifest_processor

MANIFEST_PROCESSOR_PROVIDERS: Dict[Identity, manifest.PROCESSOR_PROVIDER] = {
    Identity("calico"): CalicoManifestProcessor,
    Identity("calico", "apiserver"): CalicoApiServerManifestProcessor,
    Identity("nginx-ingress-controller"): get_ingress_nginx_manifest_processor,
    Identity("kubernetes-dashboard"): get_dashboard_manifest_processor,
    Identity("local-path-provisioner"): LocalPathProvisionerManifestProcessor,
}


def is_manifest_installed(cluster: KubernetesCluster, manifest_identity: Identity) -> bool:
    """
    Checks if the manifest is (to be) installed during the regular installation procedure.

    :param cluster: KubernetesCluster object
    :param manifest_identity: A pair of (plugin_name, manifest_id) that uniquely identifies the manifest
    :return: `true` is manifest is (to be) installed.
    """
    return get_manifest_processor(cluster, manifest_identity) is not None


def get_manifest_processor(cluster: KubernetesCluster, manifest_identity: Identity) -> Optional[manifest.Processor]:
    """
    Return actual manifest processor if it is (to be) installed during the regular installation procedure.

    :param cluster: KubernetesCluster object
    :param manifest_identity: A pair of (plugin_name, manifest_id) that uniquely identifies the manifest
    :return: actual manifest processor
    """
    if _is_manifest_disabled(cluster.inventory, manifest_identity):
        return None

    config = _get_manifest_installation_step(cluster.inventory, manifest_identity)
    if config is None:
        return None

    return _convert_config_to_manifest_processor(cluster.log, cluster.inventory, config)


def _convert_config_to_manifest_processor(logger: log.VerboseLogger, inventory: dict, config: dict) -> manifest.Processor:
    arguments = config['arguments']
    manifest_identity = Identity(arguments['plugin_name'], arguments.get('manifest_id'))

    return _get_manifest_processor(
        logger, inventory, manifest_identity,
        arguments.get('original_yaml_path'), arguments.get('destination_name'))


def _get_manifest_installation_step(inventory: dict, manifest_identity: Identity) -> Optional[dict]:
    plugin_name = manifest_identity.plugin_name
    items = inventory['plugins'][plugin_name]['installation']['procedures']
    for i, item in enumerate(items):
        if 'python' not in item:
            continue

        config: dict = item['python']
        if config['module'] != "plugins/builtin.py" or config['method'] != "apply_yaml":
            continue

        # presence of arguments and signature of apply_yaml is validated in plugins.verify_python
        arguments = config['arguments']
        declared_name = arguments['plugin_name']
        if declared_name != plugin_name:
            raise Exception(f"Unexpected 'plugin_name={declared_name}' argument "
                            f"in {plugin_name!r} installation step {i}.")

        declared_identity = Identity(plugin_name, arguments.get('manifest_id'))
        if manifest_identity != declared_identity:
            continue

        return config

    return None


def _is_manifest_disabled(inventory: dict, manifest_identity: Identity) -> bool:
    plugin_name = manifest_identity.plugin_name
    if not inventory["plugins"][plugin_name]["install"]:
        return True

    if manifest_identity == Identity("calico", "apiserver") and not calico.is_apiserver_enabled(inventory):
        return True

    return False


@enrichment(EnrichmentStage.FULL)
def verify_inventory(cluster: KubernetesCluster) -> None:
    inventory = cluster.inventory
    for manifest_identity in MANIFEST_PROCESSOR_PROVIDERS:
        if _is_manifest_disabled(inventory, manifest_identity):
            continue

        config = _get_manifest_installation_step(inventory, manifest_identity)
        if config is not None:
            processor = _convert_config_to_manifest_processor(cluster.log, inventory, config)
            processor.validate_inventory()
        else:
            cluster.log.warning(f"Invocation of plugins.builtin.apply_yaml for {manifest_identity.repr_id()} "
                                f"is not found for {manifest_identity.plugin_name!r} plugin. "
                                f"Such configuration is obsolete, and support for it may be stopped in future releases.")


def _get_manifest_processor(logger: log.VerboseLogger, inventory: dict, manifest_identity: Identity,
                            original_yaml_path: Optional[str] = None,
                            destination_name: Optional[str] = None) -> manifest.Processor:
    if manifest_identity not in MANIFEST_PROCESSOR_PROVIDERS:
        raise Exception(f"Cannot find processor of {manifest_identity.repr_id()} "
                        f"for {manifest_identity.plugin_name!r} plugin.")

    processor_provider = MANIFEST_PROCESSOR_PROVIDERS[manifest_identity]
    return processor_provider(logger, inventory, original_yaml_path, destination_name)


def apply_yaml(cluster: KubernetesCluster, plugin_name: str,
               manifest_id: Optional[str] = None,
               original_yaml_path: Optional[str] = None,
               destination_name: Optional[str] = None) -> None:
    manifest_identity = Identity(plugin_name, manifest_id)
    # Since the method is called from the inventory, the installation step is certainly present.
    # Though it can still be disabled by other inventory properties.
    if _is_manifest_disabled(cluster.inventory, manifest_identity):
        cluster.log.debug(f"Skip installing of the {manifest_identity.repr_id()} for {plugin_name!r} plugin.")
        return

    processor = _get_manifest_processor(cluster.log, cluster.inventory, manifest_identity,
                                        original_yaml_path, destination_name)

    manifest = processor.enrich()
    processor.apply(cluster, manifest)


def get_namespace_to_necessary_pss_profiles(cluster: KubernetesCluster) -> Dict[str, str]:
    """
    :param cluster: KubernetesCluster object
    :return: minimal required PSS profiles for all installed plugins' namespaces
    """
    result = {}
    for manifest_identity in MANIFEST_PROCESSOR_PROVIDERS:
        # Use _is_manifest_disabled() instead of `not is_manifest_installed()`,
        # because old inventories might still refer to .yaml.j2 templates where PSS was also managed automatically.
        if _is_manifest_disabled(cluster.inventory, manifest_identity):
            continue

        processor = _get_manifest_processor(cluster.log, cluster.inventory, manifest_identity)
        result.update(processor.get_namespace_to_necessary_pss_profiles())

    return result



# Copyright 2021-2022 NetCracker Technology Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from kubemarine.core.cluster import KubernetesCluster, EnrichmentStage, enrichment
from kubemarine.core.group import RunnersGroupResult, NodeGroup
from kubemarine.cri import docker, containerd

ERROR_FORBIDDEN_CRI_SECTION = "{key} is not used, please remove {value} config from `services.cri` section"


@enrichment(EnrichmentStage.FULL)
def enrich_inventory(cluster: KubernetesCluster) -> None:
    inventory = cluster.inventory
    cri_impl = get_cri_impl(inventory)

    if cri_impl == "docker":
        forbidden_cri_sections = {"containerd": ["containerdConfig", "containerdRegistriesConfig"]}
        del inventory['services']['cri']['containerdConfig']
    else:
        forbidden_cri_sections = {"docker": ["dockerConfig"]}
        del inventory['services']['cri']['dockerConfig']

    for key, sections in forbidden_cri_sections.items():
        for value in sections:
            if value in cluster.raw_inventory.get('services', {}).get('cri', {}):
                raise Exception(ERROR_FORBIDDEN_CRI_SECTION.format(key=key, value=value))

    # Enrich containerdConfig
    if cri_impl == "containerd":
        containerd.enrich_inventory(cluster)


@enrichment(EnrichmentStage.PROCEDURE, procedures=['upgrade'])
def enrich_upgrade_inventory(cluster: KubernetesCluster) -> None:
    cri_impl = get_cri_impl(cluster.previous_inventory)
    if cri_impl == "containerd":
        containerd.enrich_upgrade_inventory(cluster)


@enrichment(EnrichmentStage.PROCEDURE, procedures=['upgrade'])
def verify_upgrade_inventory(cluster: KubernetesCluster) -> None:
    cri_impl = get_cri_impl(cluster.previous_inventory)
    if cri_impl == "containerd":
        containerd.verify_upgrade_inventory(cluster)


def get_cri_impl(inventory: dict) -> str:
    cri_impl: str = inventory['services']['cri']['containerRuntime']
    return cri_impl


def install(group: NodeGroup) -> RunnersGroupResult:
    cri_impl = group.cluster.inventory['services']['cri']['containerRuntime']

    if cri_impl == "docker":
        return docker.install(group)
    else:
        return containerd.install(group)


def configure(group: NodeGroup) -> RunnersGroupResult:
    cri_impl = group.cluster.inventory['services']['cri']['containerRuntime']

    if cri_impl == "docker":
        return docker.configure(group)
    else:
        return containerd.configure(group)


def prune(group: NodeGroup, all_implementations: bool = False) -> RunnersGroupResult:
    cri_impl = group.cluster.inventory['services']['cri']['containerRuntime']

    result = RunnersGroupResult(group.cluster, {})
    if cri_impl == "docker" or all_implementations:
        result = docker.prune(group)

    if cri_impl == "containerd" or all_implementations:
        result = containerd.prune(group)

    return result
